@@ -12,7 +12,7 @@ from . import packet
 from . import payload
 from . import socket
 
-default_logger = logging.getLogger('engineio')
+default_logger = logging.getLogger('engineio.server')
 
 
 class Server(object):
@@ -28,7 +28,7 @@ class Server(object):
                        argument is not given, "eventlet" is tried first, then
                        "gevent_uwsgi", then "gevent", and finally "threading".
                        The first async mode that has all its dependencies
-                       installed is then one that is chosen.
+                       installed is the one that is chosen.
     :param ping_timeout: The time in seconds that the client waits for the
                          server to respond before disconnecting. The default
                          is 60 seconds.
@@ -117,7 +117,7 @@ class Server(object):
         for mode in modes:
             try:
                 self._async = importlib.import_module(
-                    'engineio.async_' + mode)._async
+                    'engineio.async_drivers.' + mode)._async
                 asyncio_based = self._async['asyncio'] \
                     if 'asyncio' in self._async else False
                 if asyncio_based != self.is_asyncio_based():
@@ -207,6 +207,65 @@ class Server(object):
             return
         socket.send(packet.Packet(packet.MESSAGE, data=data, binary=binary))
 
+    def get_session(self, sid):
+        """Return the user session for a client.
+
+        :param sid: The session id of the client.
+
+        The return value is a dictionary. Modifications made to this
+        dictionary are not guaranteed to be preserved unless
+        ``save_session()`` is called, or when the ``session`` context manager
+        is used.
+        """
+        socket = self._get_socket(sid)
+        return socket.session
+
+    def save_session(self, sid, session):
+        """Store the user session for a client.
+
+        :param sid: The session id of the client.
+        :param session: The session dictionary.
+        """
+        socket = self._get_socket(sid)
+        socket.session = session
+
+    def session(self, sid):
+        """Return the user session for a client with context manager syntax.
+
+        :param sid: The session id of the client.
+
+        This is a context manager that returns the user session dictionary for
+        the client. Any changes that are made to this dictionary inside the
+        context manager block are saved back to the session. Example usage::
+
+            @eio.on('connect')
+            def on_connect(sid, environ):
+                username = authenticate_user(environ)
+                if not username:
+                    return False
+                with eio.session(sid) as session:
+                    session['username'] = username
+
+            @eio.on('message')
+            def on_message(sid, msg):
+                with eio.session(sid) as session:
+                    print('received message from ', session['username'])
+        """
+        class _session_context_manager(object):
+            def __init__(self, server, sid):
+                self.server = server
+                self.sid = sid
+                self.session = None
+
+            def __enter__(self):
+                self.session = self.server.get_session(sid)
+                return self.session
+
+            def __exit__(self, *args):
+                self.server.save_session(sid, self.session)
+
+        return _session_context_manager(self, sid)
+
     def disconnect(self, sid=None):
         """Disconnect a client.
 
